#   -------------------------------------------------------------
#   Licensed under the MIT License. See LICENSE in project root for information.
#   -------------------------------------------------------------
"""lingualabpy"""
from __future__ import annotations

__version__ = "0.0.4"

default_config = {
    "participant_col": "participant_id",
    "participant_label": "IE",
    "clinician_label": "IV",
    "f0_bounds": {
        "female": [100.0, 600.0],
        "male": [75.0, 300.0],
    },
    "unit_frequency": "Hertz",
}

from lingualabpy.io import read_audio, read_docx, read_json, write_json, read_textgrid



from collections import defaultdict
import numpy as np
from parselmouth import Sound
from parselmouth.praat import call

from lingualabpy.tools.data import UnchangeableDict


def measure_pitch(sound: Sound, f0min: str, f0max: str, unit: str) -> UnchangeableDict:
    """
    This function measures duration, pitch, HNR, jitter, and shimmer
    This is the function to measure source acoustics using default male parameters.
    """
    # compute usefull praat object
    pitch = call(sound, "To Pitch", 0.0, f0min, f0max)
    harmonicity = call(sound, "To Harmonicity (cc)", 0.01, f0min, 0.1, 1.0)
    point_process = call(sound, "To PointProcess (periodic, cc)", f0min, f0max)

    # metrics container
    metrics = UnchangeableDict()

    # Metrics computation
    metrics["duration"] = call(sound, "Get total duration")
    metrics["f0_mean"] = call(pitch, "Get mean", 0, 0, unit)
    metrics["F0_std"] = call(pitch, "Get standard deviation", 0, 0, unit)
    metrics["hnr"] = call(harmonicity, "Get mean", 0, 0)

    # jitter
    jitter_types = ["local", ["local", "absolute"], "rap", "ppq5", "ddp"]
    for jitter_type in jitter_types:
        if isinstance(jitter_type, list):
            metric_name = f"jitter_{'_'.join(jitter_type)}"
            praat_function = f"Get jitter ({', '.join(jitter_type)})"
        else:
            metric_name = f"jitter_{jitter_type}"
            praat_function = f"Get jitter ({jitter_type})"
        metrics[metric_name] = call(
            point_process, praat_function, 0, 0, 0.0001, 0.02, 1.3
        )

    # shimmer
    shimmer_types = ["local", "local_dB", "apq3", "apq5", "apq11", "dda"]
    for shimmer_type in shimmer_types:
        metric_name = f"shimmer_{shimmer_type}"
        praat_function = f"Get shimmer ({shimmer_type})"
        metrics[metric_name] = call(
            [sound, point_process], praat_function, 0, 0, 0.0001, 0.02, 1.3, 1.6
        )

    return metrics


def measure_formants(
    sound: Sound, f0min: str, f0max: str, unit: str
) -> UnchangeableDict:
    """
    This function measures formants at each glottal pulse

    Puts, D. A., Apicella, C. L., & Cárdenas, R. A. (2012). Masculine voices signal men's threat potential in forager and industrial societies. Proceedings of the Royal Society of London B: Biological Sciences, 279(1728), 601-609.

    Adapted from: DOI 10.17605/OSF.IO/K2BHS
    """
    # compute usefull praat object
    point_process = call(sound, "To PointProcess (periodic, cc)", f0min, f0max)
    formants = call(sound, "To Formant (burg)", 0.0025, 5, 5000, 0.025, 50)
    number_of_points = call(point_process, "Get number of points")

    # metrics container
    metrics = UnchangeableDict()

    # Measure formants only at glottal pulses
    formants_list = defaultdict(list)
    for index in range(1, number_of_points + 1):
        time = call(point_process, "Get time from index", index)
        for pulse in [1, 2, 3, 4]:
            value = call(formants, "Get value at time", pulse, time, unit, "Linear")
            if str(value) != "nan":
                formants_list[pulse].append(value)

    # calculate mean and median formants across pulses, median is what is used in all subsequent calculations
    for pulse in [1, 2, 3, 4]:
        metrics[f"formants_{pulse}_mean"] = np.mean(formants_list[pulse])
        metrics[f"formants_{pulse}_median"] = np.median(formants_list[pulse])

    return metrics



from textgrids import Interval


def interval_to_list(interval: Interval) -> list[float]:
    """"""
    return [interval.xmin, interval.xmax]


def is_overlap(interval0: Interval, interval1: Interval) -> bool:
    """Check if two intervals overlap"""
    return interval0.xmin <= interval1.xmax and interval1.xmin <= interval0.xmax


def remove_overlap(interval: Interval, interval_to_remove: Interval) -> list[Interval]:
    """"""
    # Return interval as a list if there is no overlap
    if not is_overlap(interval, interval_to_remove):
        return [interval]

    else:
        updated_intervals = []

        # If the start of the interval is before the start of the interval to be removed,
        # add the non-overlapping part to the result.
        if interval.xmin < interval_to_remove.xmin:
            updated_intervals.append(
                Interval(xmin=interval.xmin, xmax=interval_to_remove.xmin)
            )

        # If the end of the interval is after the end of the interval to be removed,
        # add the non-overlapping part to the result.
        if interval.xmax > interval_to_remove.xmax:
            updated_intervals.append(
                Interval(xmin=interval_to_remove.xmax, xmax=interval.xmax)
            )

        return updated_intervals


def intervals_masking(
    intervals: list[Interval], intervals_mask: list[Interval]
) -> list[list[float]]:
    """"""
    # Each intervals mask will be remove from all the intervals
    for interval_to_remove in intervals_mask:
        new_intervals = []
        for interval in intervals:

            # if the start of the interval is after the end of the mask
            # we can just add the interval they are sorted
            if interval.xmin > interval_to_remove.xmax:
                new_intervals.append(interval)

            else:
                new_intervals += remove_overlap(interval, interval_to_remove)

        intervals = new_intervals

    return [interval_to_list(_) for _ in intervals]



from collections import UserDict
from pandas import DataFrame

from typing import Any, Dict, List


class UnchangeableDict(UserDict):
    """A dictionary in which you can add new keys but not modify them in the future."""

    def __setitem__(self, key: Any, item: Any) -> None:
        try:
            self.__getitem__(key)
            raise ValueError("duplicate key '{}' found".format(key))
        except KeyError:
            return super().__setitem__(key, item)


def merge_participants_to_df(
    data_participants: List[Dict[Any, Any]],
    participant_col: str,
) -> DataFrame:
    # Check if all data have a `participant_col` key
    participant_col_checks = [_.get(participant_col) for _ in data_participants]
    if not all(participant_col_checks):
        raise Exception(
            f"One of the samples does not contain the '{participant_col}' information."
        )

    # Check if there are no duplicates in the data
    df_raw = DataFrame.from_dict(data_participants)
    df_melt = df_raw.melt(id_vars=[participant_col]).dropna()
    df_for_test = df_melt.drop(columns="value")
    duplicates = df_for_test[df_for_test.duplicated()]

    if duplicates.empty:
        return df_melt.pivot(index=participant_col, columns="variable")["value"]
    else:
        error_msg = "There are duplicates in your data "
        for participant_id, variable in duplicates.values:
            error_msg += f"\n{participant_id}: {variable}"
        raise Exception(error_msg)



import click

from lingualabpy import read_docx, write_json
from lingualabpy.text.parser import parse_waywithwords


@click.command()
@click.option("--origin", default="waywithwords", show_default=True)
@click.argument("docx_path", nargs=1, type=click.Path(exists=True))
@click.argument("output", nargs=1)
def main(docx_path, output, origin):
    """Doc"""
    document = read_docx(docx_path)

    if origin == "waywithwords":
        data = parse_waywithwords(document)

    else:
        raise ValueError(f"{origin} is not implemented")

    write_json(data, output)



import re
from textgrids import TextGrid, Interval
import warnings


def extract_intervals(textgrid: TextGrid, speakers: list[str]) -> list[list[Interval]]:
    """"""
    # Check if speakers are in the textgrid tiers
    tiers = set(textgrid.keys())
    if not set(speakers).issubset(tiers):
        raise ValueError(
            f"Some speaker(s) '{speakers}' are not a tier in the TextGrid '{tiers}'"
        )

    # Check if there is other speaker in the textgrid
    if not set(speakers) == tiers:
        warnings.warn(
            f"TextGrid '{tiers}' have more speakers than specify '{speakers}'"
        )

    # Extraction of intervals with text value
    speakers_intervals = []
    for speaker in speakers:
        speaker_intervals = []
        for interval in textgrid[speaker]:
            # Cleaning of the interval text
            interval.text = (
                interval.text.encode().decode("unicode_escape").strip(" \n\r\t")
            )
            if interval.text:
                speaker_intervals.append(interval)
        speakers_intervals.append(speaker_intervals)

    # Checking if all intervals are correctly labeled
    def interval_qc(intervals, label):
        labels = set([_.text for _ in intervals])
        if not (len(labels) == 1 and labels.pop() == label):
            raise Exception(
                f"TextGrid was not labeled correctly, current label(s) '{labels}', should be '{label}'."
            )

    for intervals, speaker in zip(speakers_intervals, speakers):
        interval_qc(intervals, speaker)

    return speakers_intervals



import click

from lingualabpy import default_config, read_audio, read_textgrid
from lingualabpy.audio.triming import extract_audio
from lingualabpy.text.textgrid import extract_intervals
from lingualabpy.tools.interval import intervals_masking


@click.command()
@click.option(
    "--participant_label",
    default=default_config["participant_label"],
    show_default=True,
)
@click.option(
    "--clinician_label",
    default=default_config["clinician_label"],
    show_default=True,
)
@click.argument("textgrid", nargs=1, type=click.Path(exists=True))
@click.argument("audiofile", nargs=1, type=click.Path(exists=True))
@click.argument("output", nargs=1)
def main(participant_label, clinician_label, textgrid, audiofile, output):
    """Doc"""
    grid = read_textgrid(textgrid)

    try:
        participant_intervals, clinician_intervals = extract_intervals(
            grid, [participant_label, clinician_label]
        )
    except Exception as e:
        raise Exception(f"Failed to extract intervals for {textgrid}", repr(e))

    participant_intervals_clean = intervals_masking(
        participant_intervals, clinician_intervals
    )

    audio = read_audio(audiofile)

    audio_clean = extract_audio(audio, participant_intervals_clean)

    audio_clean.export(output)



import click
import json

from lingualabpy import default_config, read_json
from lingualabpy.tools.data import merge_participants_to_df


@click.command()
@click.option(
    "-c",
    "--column",
    default=default_config["participant_col"],
    show_default=True,
)
@click.argument("jsons", nargs=-1, type=click.Path(exists=True))
@click.argument("output", nargs=1)
def main(column, jsons, output):
    """ """
    data = []
    for path in jsons:
        data.append(read_json(path))
    df = merge_participants_to_df(data, participant_col=column)
    df.to_csv(output)



import re
from docx import Document
from collections import defaultdict


def parse_waywithwords(document: Document) -> dict:
    """"""
    waywithwords = {
        "IV": "interviewer",
        "IE": "interviewee",
    }

    results = defaultdict(list)

    for para in document.paragraphs:
        try:
            content = para.text.split()
            speaker = content[0]
            transcription = " ".join(content[1:])
        except:
            speaker = None

        if (
            speaker in waywithwords.keys()
            and not transcription.lower() in waywithwords.values()
        ):
            results[waywithwords[speaker]].append(transcription)

        elif re.findall(r"[0-9][0-9]:[0-5][0-9]:[0-5][0-9]", para.text):
            results["time"].append(para.text)

        else:
            results["remainder"].append(para.text)

    return results



"""
Module contains tools for processing files
"""

import json
from docx import Document
from textgrids import TextGrid
from pydub import AudioSegment

from typing import Union


# audio files
def read_audio(sound_path: str) -> AudioSegment:
    """"""
    return AudioSegment.from_file(sound_path)


# .docx files
def read_docx(docx_path: str) -> Document:
    """"""
    return Document(docx_path)


# .json files
def read_json(json_path: str) -> Union[list, dict]:
    """"""
    with open(json_path, "r") as file:
        content = json.load(file)
    return content


def write_json(data: Union[list, dict], json_path: str) -> None:
    """"""
    with open(json_path, "w") as file:
        json.dump(data, file, indent=4)


# .TextGrid files
def read_textgrid(textgrid_path: str) -> TextGrid:
    """"""
    return TextGrid(textgrid_path)



from pydub import AudioSegment


def extract_audio(audio: AudioSegment, intervals: list[list[float]]) -> AudioSegment:
    """"""
    new_audio = AudioSegment.empty()

    for start, end in intervals:
        new_audio += audio[start * 1000 : end * 1000]

    return new_audio



import click
from parselmouth import Sound
from pathlib import Path

from lingualabpy import default_config, write_json
from lingualabpy.audio.metrics import measure_pitch, measure_formants


@click.command()
@click.option(
    "--sex",
    type=click.Choice(["female", "male"]),
    help=f"Set f0min and f0max for praat analysis. {default_config['f0_bounds']}",
)
@click.option(
    "--f0min",
    type=float,
    help="Define f0min for praat analysis. Not required if sex is specify",
)
@click.option(
    "--f0max",
    type=float,
    help="Define f0max for praat analysis. Not required if sex is specify",
)
@click.option(
    "--unit_frequency",
    default=default_config["unit_frequency"],
    show_default=True,
)
@click.option("--participant_id", "-p", default=None, help="")
@click.option("--output_json", default=None, help="")
@click.argument("audiofile", nargs=1, type=click.Path(exists=True))
def main(sex, f0min, f0max, unit_frequency, participant_id, output_json, audiofile):
    """Doc"""
    if sex:
        f0min, f0max = default_config["f0_bounds"][sex]
    else:
        if not f0min or not f0max:
            raise click.UsageError(
                "'--f0min' and '--f0max' are required if '--sex' is not specified"
            )

    sound = Sound(audiofile)
    metrics = measure_pitch(sound, f0min, f0max, unit_frequency)
    metrics.update(measure_formants(sound, f0min, f0max, unit_frequency))

    audiofile_stem = Path(audiofile).stem

    if participant_id:
        metrics["participant_id"] = participant_id
    else:
        metrics["participant_id"] = audiofile_stem.split("_")[0]

    if not output_json:
        output_json = audiofile_stem + "_metric-audio.json"

    write_json(dict(metrics), output_json)
